from ._sdl import sdl2
from .Window import Window, InternalWindow
from .Image import icon_to_image
from .Common import _temp, _calculate_object2d
from .Camera import Camera
from .Material import Material, DefaultMaterial
from .Color import Color
from .Time import Time, fps_to_time
from .Constants import Pivot2D
from .Mouse import MouseButton, simpleHover, object2dSimpleHover
from .Text import _ObjectText, _DefaultFont

__all__ = ["Button", "ButtonText", "ButtonImage", "ButtonIcon"]

class _button:
    def __init__(self, location=(0, 0), rotation=0, size=(0, 0), scale=(1, 1)):
        self._size = list(size)
        self._location = list(location)
        self._rotation = rotation
        self._scale = list(scale)
        self._pivot = Pivot2D.TopLeftSide
        self._cursor = 11

    def hover(self, window: Window | InternalWindow, camera: Camera):
        return object2dSimpleHover(window, camera, self._location, self._size, self._scale, self._pivot)

    def button(self, button: int, window, camera: Camera, multiple_click: bool = False) -> bool:
        #self.button_active = True
        if self.hover(window, camera):
            _temp.MouseCursor = self._cursor
            if MouseButton(button, multiple_click):
                return True
        return False

    @property
    def location(self) -> list[int, int]:
        return self._location

    @location.setter
    def location(self, location):
        self._location = location

    @property
    def size(self) -> list[int, int]:
        return self._size

    @size.setter
    def size(self, size):
        self._size = size

    @property
    def rotation(self):
        return self._rotation

    @rotation.setter
    def rotation(self, rotation: int | float):
        if self._rotation != rotation:
            self._rotation = round(rotation, 4)
            self._rotation %= 360

    @property
    def scale(self):
        return self._scale

    @scale.setter
    def scale(self, scale):
        self._scale = scale

    @property
    def pivot(self):
        return self._pivot

    @pivot.setter
    def pivot(self, pivot):
        self._pivot = pivot

    def cursor(self, cursor: int):
        self._cursor = cursor

class Button(_button):
    def __init__(self, location=(0, 0), rotation=0, size=(0, 0), scale=(1, 1), material=DefaultMaterial, material_hover=DefaultMaterial):
        super().__init__(location, rotation, size, scale)
        self._border_size = 0
        self._border_color = Color((100, 100, 255))
        self._border_radius = [0, 0, 0, 0]

        self._material = material
        self._material_hover = material_hover

        self.motion_tick_time = {"x": Time(fps_to_time(60)), "y": Time(fps_to_time(60))}

        self.variables = {}

        self._showMoreDetailsOfCollisions = False

        self.object_render = True
        self.thed_render = self.always_render = False

        self.cache_object = None
        self.cache_object_hover = None
        self.cache_object_tx = None
        self.cache_object_tx_hover = None

        self.button_active = False

    def render(self, window):
        if self.object_render or self.always_render:
            if len(self._material.textures) > 0:
                if self.cache_object is None:
                    self._material.update()
                    self.cache_object = self._material.surface
                    if self.cache_object_tx is not None:
                        sdl2.SDL_DestroyTexture(self.cache_object_tx)
                    self.cache_object_tx = sdl2.SDL_CreateTextureFromSurface(window.renderer, self.cache_object).contents
            else:
                if self.cache_object is not None:
                    self.cache_object = None
                    if self.cache_object_tx is not None:
                        sdl2.SDL_DestroyTexture(self.cache_object_tx)
                        self.cache_object_tx = None
                self._material.update()

            if len(self._material_hover.textures) > 0:
                if self.cache_object_hover is None:
                    self._material_hover.update()
                    self.cache_object_hover = self._material_hover.surface
                    if self.cache_object_tx_hover is not None:
                        sdl2.SDL_DestroyTexture(self.cache_object_tx_hover)
                    self.cache_object_tx_hover = sdl2.SDL_CreateTextureFromSurface(window.renderer, self.cache_object_hover).contents
            else:
                if self.cache_object_hover is not None:
                    self.cache_object_hover = None
                    if self.cache_object_tx_hover is not None:
                        sdl2.SDL_DestroyTexture(self.cache_object_tx_hover)
                        self.cache_object_tx_hover = None
                self._material_hover.update()
            self.object_render = False

    def drawObject(self, window, camera: Camera = None):
        if window.__WindowActive__:
            if self not in window.drawnObjects:
                render, cache_location, cache_size = _calculate_object2d(self._location, self._size, self._rotation, self._scale, window, camera, self._pivot)
                if render:
                    window.drawnObjects.append(self)

                    self.render(window)

                    _material, _cache_object, _cache_object_tx = (self._material_hover, self.cache_object_hover, self.cache_object_tx_hover) if simpleHover(window, cache_location, cache_size) else (self._material, self.cache_object, self.cache_object_tx)

                    if _cache_object is None:
                        window.drawSquare(cache_location, cache_size, self._rotation, self._border_radius[0], _material.color)
                    else:
                        _material.update()
                        _cache_object = _material.surface
                        sdl2.SDL_UpdateTexture(_cache_object_tx, None, _cache_object.pixels, _cache_object.pitch)
                        window.blit(_cache_object_tx, cache_location, cache_size, self._rotation)
                    if self._border_size != 0:
                        window.drawEdgesSquare(cache_location, cache_size, self._rotation, self._border_size, self._border_radius[0], self._border_color)

    def cleanCache(self):
        self.cache_object = None
        self.cache_object_hover = None
        if self.cache_object_tx is not None:
            sdl2.SDL_DestroyTexture(self.cache_object_tx)
            self.cache_object_tx = None
        if self.cache_object_tx_hover is not None:
            sdl2.SDL_DestroyTexture(self.cache_object_tx_hover)
            self.cache_object_tx_hover = None

    def close(self):
        self.cleanCache()
        del self

    @property
    def material(self):
        return self._material

    @material.setter
    def material(self, material: Material):
        if self._material != material:
            self._material = material
            self.object_render = False

    @property
    def materialHover(self):
        return self._material_hover

    @materialHover.setter
    def materialHover(self, material: Material):
        if self._material_hover != material:
            self._material_hover = material
            self.object_render = False

class ButtonText(_button, _ObjectText):
    def __init__(self, location=(0, 0), rotation=0, size=(0, 0), scale=(1, 1), font_size: int = 20, text: str = "", font=_DefaultFont, material=DefaultMaterial, material_hover=DefaultMaterial):
        super().__init__(location, rotation, size, scale)
        super(_button, self).__init__(font_size, text, font)

        self._border_size = 0
        self._border_color = Color((100, 100, 255))
        self._border_radius = [0, 0, 0, 0]

        self._material = material
        self._material_hover = material_hover
        self.materials = [self._material, self._material_hover]

        self.motion_tick_time = {"x": Time(fps_to_time(60)), "y": Time(fps_to_time(60))}

        self.variables = {}

        self._showMoreDetailsOfCollisions = False

        self.object_render = True
        self.thed_render = self.always_render = False

        self.cache_object = None
        self.cache_object_hover = None
        self.cache_object_tx = None
        self.cache_object_tx_hover = None

        self.button_active = False

    def render(self, window):
        if self.object_render or self.always_render:
            self._render()
            if len(self._material.textures) > 0:
                if self.cache_object is None:
                    self._material.update()
                    self.cache_object = self._material.surface
                    if self.cache_object_tx is not None:
                        sdl2.SDL_DestroyTexture(self.cache_object_tx)
                    self.cache_object_tx = sdl2.SDL_CreateTextureFromSurface(window.renderer, self.cache_object).contents
            else:
                if self.cache_object is not None:
                    self.cache_object = None
                    if self.cache_object_tx is not None:
                        sdl2.SDL_DestroyTexture(self.cache_object_tx)
                        self.cache_object_tx = None
                self._material.update()

            if len(self._material_hover.textures) > 0:
                if self.cache_object_hover is None:
                    self._material_hover.update()
                    self.cache_object_hover = self._material_hover.surface
                    if self.cache_object_tx_hover is not None:
                        sdl2.SDL_DestroyTexture(self.cache_object_tx_hover)
                    self.cache_object_tx_hover = sdl2.SDL_CreateTextureFromSurface(window.renderer, self.cache_object_hover).contents
            else:
                if self.cache_object_hover is not None:
                    self.cache_object_hover = None
                    if self.cache_object_tx_hover is not None:
                        sdl2.SDL_DestroyTexture(self.cache_object_tx_hover)
                        self.cache_object_tx_hover = None
                self._material_hover.update()
            self.object_render = False

    def drawObject(self, window, camera: Camera = None, render=True):
        if window.__WindowActive__:
            if self not in window.drawnObjects:
                window.drawnObjects.append(self)
                _render, cache_location, cache_size = _calculate_object2d(self._location, self._size, self._rotation, self._scale, window, camera, self._pivot)
                if _render:

                    self.render(window)

                    _material, _cache_object, _cache_object_tx = (self._material_hover, self.cache_object_hover, self.cache_object_tx_hover) if simpleHover(window, cache_location, cache_size) else (self._material, self.cache_object, self.cache_object_tx)

                    if _cache_object is None:
                        window.drawSquare(cache_location, cache_size, self._rotation, self._border_radius[0], _material.color)
                    else:
                        _material.update()
                        _cache_object = _material.surface
                        sdl2.SDL_UpdateTexture(_cache_object_tx, None, _cache_object.pixels, _cache_object.pitch)
                        window.blit(_cache_object_tx, cache_location, cache_size, self._rotation)
                    if render:
                        self._render()
                    window.blit(self._surface, [cache_location[0] + (cache_size[0] // 2) - (self.surfaceSize[0] // 2), cache_location[1] + (cache_size[1] // 2) - (self.surfaceSize[1] // 2)], None, self._rotation)
                    if self._border_size != 0:
                        window.drawEdgesSquare(cache_location, cache_size, self._rotation, self._border_size, self._border_radius[0], self._border_color)

    def cleanCache(self):
        self.cache_object = None
        self.cache_object_hover = None
        if self.cache_object_tx is not None:
            sdl2.SDL_DestroyTexture(self.cache_object_tx)
            self.cache_object_tx = None
        if self.cache_object_tx_hover is not None:
            sdl2.SDL_DestroyTexture(self.cache_object_tx_hover)
            self.cache_object_tx_hover = None

    def close(self):
        self.cleanCache()
        del self

    @property
    def material(self):
        return self._material

    @material.setter
    def material(self, material: Material):
        if self._material != material:
            self._material = material
            self.object_render = True

    @property
    def materialHover(self):
        return self._material_hover

    @materialHover.setter
    def materialHover(self, material: Material):
        if self._material_hover != material:
            self._material_hover = material
            self.object_render = True

class ButtonImage(_button):
    def __init__(self, location=(0, 0), rotation=0, size=(0, 0), scale=(1, 1), image=None, image_hover=None):
        super().__init__(location, rotation, size, scale)
        self._border_size = 0
        self._border_color = Color((100, 100, 255))
        self._border_radius = [0, 0, 0, 0]

        self._image = image
        self._image_hover = image_hover

        self.motion_tick_time = {"x": Time(fps_to_time(60)), "y": Time(fps_to_time(60))}

        self.variables = {}

        self._showMoreDetailsOfCollisions = False

        self.object_render = self.thed_render = self.always_render = False

        self.cache_object = None
        self.cache_object_hover = None
        self.cache_object_tx = None
        self.cache_object_tx_hover = None

        self.button_active = False

    def render(self, window):
        if not self.object_render:
            if self.cache_object is None:
                self.cache_object = self._image.images[0]
                if self.cache_object_tx is not None:
                    sdl2.SDL_DestroyTexture(self.cache_object_tx)
                self.cache_object_tx = sdl2.SDL_CreateTextureFromSurface(window.renderer, self.cache_object).contents
                sdl2.SDL_SetTextureScaleMode(self.cache_object_tx, 1)
            if self.cache_object_hover is None:
                self.cache_object_hover = self._image_hover.images[0]
                if self.cache_object_tx_hover is not None:
                    sdl2.SDL_DestroyTexture(self.cache_object_tx_hover)
                self.cache_object_tx_hover = sdl2.SDL_CreateTextureFromSurface(window.renderer, self.cache_object_hover).contents
                sdl2.SDL_SetTextureScaleMode(self.cache_object_tx_hover, 1)
        self.object_render = True

    def drawObject(self, window, camera: Camera = None):
        if window.__WindowActive__:
            if self not in window.drawnObjects:
                render, cache_location, cache_size = _calculate_object2d(self._location, self._size, self._rotation, self._scale, window, camera, self._pivot)
                if render:
                    window.drawnObjects.append(self)
                    if not self.object_render or self.always_render or window.render_all_objects:
                        self.object_render = False
                        self.render(window)

                    _cache_object_tx = self.cache_object_tx_hover if simpleHover(window, cache_location, cache_size) else self.cache_object_tx

                    window.blit(_cache_object_tx, cache_location, cache_size, self._rotation)
                    if self._border_size != 0:
                        window.drawEdgesSquare(cache_location, cache_size, self._rotation, self._border_size, self._border_radius[0], self._border_color)

    def cleanCache(self):
        self.cache_object = None
        self.cache_object_hover = None
        if self.cache_object_tx is not None:
            sdl2.SDL_DestroyTexture(self.cache_object_tx)
            self.cache_object_tx = None
        if self.cache_object_tx_hover is not None:
            sdl2.SDL_DestroyTexture(self.cache_object_tx_hover)
            self.cache_object_tx_hover = None

    def close(self):
        self.cleanCache()
        del self

    @property
    def image(self):
        return self._image

    @image.setter
    def image(self, image: Material):
        if self._image != image:
            self._image = image
            self.object_render = False

    @property
    def imageHover(self):
        return self._image_hover

    @imageHover.setter
    def imageHover(self, image: Material):
        if self._image_hover != image:
            self._image_hover = image
            self.object_render = False

class ButtonIcon(ButtonImage):
    def __init__(self, location=(0, 0), rotation=0, size=(0, 0), scale=(1, 1), icon=None, icon_hover=None):
        super().__init__(location, rotation, size, scale, icon_to_image(icon), icon_to_image(icon_hover))



import os
from ctypes import CDLL
from ..Log import LogError, LogCritical

def nullFunction(*args):
    return

def get_depsPath(file_name: str | list):
    file_name = file_name if type(file_name) == str else file_name[0]
    if os.path.exists(f"{os.path.abspath('./')}/{file_name}"):
        dll_path = os.path.abspath(f'./{file_name}')
    elif os.path.exists(f"{os.path.abspath(os.path.dirname(__file__))}/{file_name}"):
        dll_path = f"{os.path.abspath(os.path.dirname(__file__))}/{file_name}"
    elif os.path.exists(f"{os.path.abspath(os.path.dirname(__file__))}/../{file_name}"):
        dll_path = f"{os.path.abspath(os.path.dirname(__file__))}/../{file_name}"
    else:
        LogError(f"{file_name} not found")
        return None
    return dll_path

class DLL(object):
    def __init__(self, path=None):
        self._dll = None if path is None else CDLL(path)

    def bindFunction(self, func_name, args=None, returns=None):
        if self._dll is None:
            return None if func_name is None else nullFunction
        else:
            if func_name is None:
                return nullFunction
            func = getattr(self._dll, func_name, None)
            if not func:
                LogError(f"{func_name}")
                return nullFunction
            func.argtypes, func.restype = args, returns
            return func

SDLFunc = DLL(get_depsPath("SDL2.dll")).bindFunction
GFXFunc = DLL(get_depsPath("SDL2_gfx.dll")).bindFunction
IMAGEFunc = DLL(get_depsPath("SDL2_image.dll")).bindFunction
MIXERFunc = DLL(get_depsPath("SDL2_mixer.dll")).bindFunction
TTFFunc = DLL(get_depsPath("SDL2_ttf.dll")).bindFunction

if None in (SDLFunc(None), GFXFunc(None), IMAGEFunc(None), MIXERFunc(None), TTFFunc(None)):
    LogCritical(f"Error when importing SDL2 libraries", "MGE dependencies not found. Run 'MGE deps install' to install them")



from .Image import Image
from ._sdl import sdl2
from .Time import Time
from .Log import LogError

__all__ = ["Texture"]

def _get_sdl2_texture_size(texture):
    ret = sdl2.SDL_QueryTexture(texture)
    if ret is None:
        ret = [0, 0, 0, 0]
        LogError("getting texture attributes")
    return [ret[2], ret[3]]

class Texture:
    def __init__(self, image: Image = None):
        self._image = None
        if isinstance(image, Image):
            self._image = image
            self._time = Time(self._image.delays[0] / 1000)
            self._time_frame = 0
            self._frame = 0
        self._tx = {}
        self._alpha = 255
        self._blurr = 0
        self._RenderScaleQuality = 1
        self._render = False

    def render(self, renderer, force=False):
        if not self._render or force:
            if id(renderer) in self._tx:
                for tx in self._tx[id(renderer)]:
                    sdl2.SDL_DestroyTexture(tx)
            self._tx[id(renderer)] = []
            self._tx[id(renderer)].clear()
            if self._image.count == 1:
                self._tx[id(renderer)] = [sdl2.SDL_CreateTextureFromSurface(renderer, self._image.images[0]).contents]
            else:
                for num in range(self._image.count):
                    self._tx[id(renderer)].append(sdl2.SDL_CreateTextureFromSurface(renderer, self._image.images[num]).contents)
            self._UpdateScaleMode()
            self._render = True
        return self.tx

    def tx(self, renderer):
        if id(renderer) not in self._tx or not self._render:
            self.render(renderer, True)
        if len(self._tx[id(renderer)]) == 1:
            return self._tx[id(renderer)][0]
        self._time_frame += self._time.tickMotion()
        if self._time_frame > 1:
            _time_int = int(self._time_frame)
            self._time_frame -= _time_int

            self._frame += _time_int
            if self._frame >= self._image.count:
                self._frame = self._frame % self._image.count

            self._time.delta_time = self._image.delays[self._frame] / 1000
        return self._tx[id(renderer)][self._image.count - 1 if self._frame >= self._image.count else self._frame]

    def _UpdateScaleMode(self):
        for textures in self._tx.values():
            for texture in textures:
                sdl2.SDL_SetTextureScaleMode(texture, self._RenderScaleQuality)

    @property
    def count(self):
        return self._image.count

    @property
    def image(self):
        return self._image

    @image.setter
    def image(self, image: Image = None):
        if isinstance(image, Image):
            self._image = image
            self._render = False

    @property
    def alpha(self):
        return self._alpha

    @alpha.setter
    def alpha(self, alpha: int):
        self._alpha = alpha

    @property
    def blurr(self):
        return self._blurr

    @blurr.setter
    def blurr(self, blurr: int):
        self._blurr = blurr

    @property
    def RenderScaleQuality(self):
        return self._RenderScaleQuality

    @RenderScaleQuality.setter
    def RenderScaleQuality(self, quality):
        if self._RenderScaleQuality != quality:
            self._RenderScaleQuality = quality if 0 <= quality <= 1 else 1
            self._UpdateScaleMode()




import os
from .Color import Color
from .Constants import Colors
from ._sdl import sdl2

__all__ = ["Controller",
           "AllowBackgroundControllerEvents"]

class Controller:
    def __init__(self, index=0):
        self._controller = sdl2.SDL_GameController()
        self._name = ""
        self._led_color = Colors.Blue
        self._index = index
        self.buttons_cache = [False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False]
        if sdl2.SDL_NumJoysticks() >= index + 1:
            self._controller = sdl2.SDL_GameControllerOpen(index)
            self._name = sdl2.SDL_GameControllerName(self._controller).decode()
            sdl2.SDL_GameControllerSetPlayerIndex(self._controller, 3)
            sdl2.SDL_GameControllerSetLED(self._controller, 0, 0, 255)

    def reload(self, index=-1):
        if sdl2.SDL_NumJoysticks() >= self._index + 1 if index == -1 else 1:
            self._controller = sdl2.SDL_GameControllerOpen(self._index if index == -1 or not type(index) == int else index)
            self._name = sdl2.SDL_GameControllerName(self._controller).decode()
            self._index = self._index if index == -1 or not type(index) == int else index

    def led(self, color: Color | bool):
        if isinstance(color, bool):
            if not color:
                sdl2.SDL_GameControllerSetLED(self._controller, 0, 0, 0)
                return
            color = self._led_color
        if isinstance(color, Color) and not color.RGB == (0, 0, 0):
            self._led_color = color
        sdl2.SDL_GameControllerSetLED(self._controller, *color.RGB)

    def button(self, button=0, multiple_click=False):
        buttons = []
        for num in range(21):
            if button == -1:
                buttons.append(bool(sdl2.SDL_GameControllerGetButton(self._controller, num)))
            else:
                if button == num:
                    if sdl2.SDL_GameControllerGetButton(self._controller, button):
                        if not self.buttons_cache[button] or multiple_click:
                            self.buttons_cache[button] = True
                            return True
                if not sdl2.SDL_GameControllerGetButton(self._controller, num):
                    self.buttons_cache[num] = False
        return buttons if buttons else False

    def analogStick(self, axis):
        xy = [0, 0]
        if axis == -1:
            xy = [0, 0, 0, 0]
            axis = (0, 1, 2, 3)
        elif axis == 0:
            axis = (0, 1)
        elif axis == 1:
            axis = (2, 3)
        for num in axis:
            value = sdl2.SDL_GameControllerGetAxis(self._controller, num)
            value = round(value / 32.767)
            xy[num] = value if not -50 < value < 50 else 0
        return xy

    def trigger(self, trigger):
        if trigger == 0:
            trigger = 4
        elif trigger == 1:
            trigger = 5
        value = sdl2.SDL_GameControllerGetAxis(self._controller, trigger)
        value = round(value / 32.767)
        value = value if not -5 < value < 5 else 0
        return value

def AllowBackgroundControllerEvents(value: bool):
    os.environ["SDL_JOYSTICK_ALLOW_BACKGROUND_EVENTS"] = "1" if value else "0"



from .Common import _temp, AllEvents
from .Constants import All

__all__ = ["keyboard", "KeyboardText", "KeyboardClick", "KeyboardState"]

def KeyboardText() -> str:
    for ev in AllEvents():
        if ev.type == 771:
            return ev.text.text.decode()

def KeyboardClick() -> int:
    for ev in AllEvents():
        if ev.type == 768:
            return ev.key.keysym.sym

def KeyboardState(key=All) -> bool | list:
    if key == -1:
        return _temp.KeyboardState
    if 0 < key < 500:
        return bool(_temp.KeyboardState[key])
    return False

def keyboard(key: int | list, multiple_click=False) -> bool | list:
    if key == -1:
        if any(KeyboardState(-1)):
            return True
        return False
    key = [key] if type(key) == int else key
    for k in key:
        if KeyboardState(k):
            if _temp.KeyboardCache[k] or multiple_click:
                _temp.KeyboardCache[k] = False
                return True
        else:
            _temp.KeyboardCache[k] = True



from ctypes import c_int, byref

from ._sdl import sdl2
from .Monitors import Monitors
from .Common import _temp, AllEvents, _calculate_object2d

__all__ = ["GetMousePosition", "GetMouseGlobalPosition", "SetMousePosition", "SetMouseGlobalPosition",
           "SetMouseVisibility", "SetMouseCursor", "simpleHover", "object2dSimpleHover",
           "MouseMotion", "MouseGlobalMotion", "MouseButton", "MouseScroll", "MouseState", "MouseMovement"]

def MouseState(button):
    if button == -1:
        return _temp.MouseState
    if 0 < button < 6:
        return bool(_temp.MouseState[button - 1])
    return False

def GetMousePosition():
    x, y = c_int(0), c_int(0)
    sdl2.SDL_GetMouseState(byref(x), byref(y))
    return [x.value, y.value]

def GetMouseGlobalPosition():
    x, y = c_int(0), c_int(0)
    sdl2.SDL_GetGlobalMouseState(byref(x), byref(y))
    return [x.value, y.value]

def SetMousePosition(pos):
    sdl2.SDL_WarpMouseInWindow(None, int(pos[0]), int(pos[1]))

def SetMouseGlobalPosition(pos):
    sdl2.SDL_WarpMouseGlobal(int(pos[0]), int(pos[1]))

def SetMouseVisibility(visibility=True):
    sdl2.SDL_ShowCursor(sdl2.SDL_ENABLE if visibility else sdl2.SDL_DISABLE)

def SetMouseCursor(cursor):
    _temp.MouseCursor = cursor if 0 <= cursor <= 12 else 0

def MouseMovement():
    loc = GetMousePosition()
    if _temp.Mouse["location"] != list(loc):
        if _temp.Mouse["location"] is None:
            _temp.Mouse["location"] = list(loc)
        ret = _temp.Mouse["location"][0] - loc[0], _temp.Mouse["location"][1] - loc[1]
        _temp.Mouse["location"] = loc
        return ret
    else:
        return 0, 0

def MouseMotion(axis, speed):
    _window_resolution = sdl2.SDL_GetWindowSize(sdl2.SDL_GetMouseFocus())
    _window_resolution = [10**6, 10**6] if not _window_resolution else _window_resolution
    _location = GetMousePosition()
    if axis == 1:  # x
        if _temp.mouse_motion_tick_time["x"].tick():
            _location[0] += speed
    elif axis == 2:  # y
        if _temp.mouse_motion_tick_time["y"].tick():
            _location[1] += speed
    for num in range(2):
        if _location[num] < 0:
            _location[num] = 0
        if _location[num] > _window_resolution[num] - 1:
            _location[num] = _window_resolution[num] - 1
    SetMousePosition(_location)

def MouseGlobalMotion(axis, speed):
    _window_resolution = Monitors[0].resolution
    _location = GetMouseGlobalPosition()
    if axis == 1:  # x
        if _temp.mouse_motion_tick_time["x"].tick():
            _location[0] += speed
    elif axis == 2:  # y
        if _temp.mouse_motion_tick_time["y"].tick():
            _location[1] += speed
    for num in range(2):
        if _location[num] < 0:
            _location[num] = 0
        if _location[num] > _window_resolution[num] - 1:
            _location[num] = _window_resolution[num] - 1
    SetMouseGlobalPosition(_location)

def MouseButton(button=1, multiple_click=False):
    for num in range(5):
        if button - 1 == num:
            if MouseState(-1)[num]:
                if not _temp.Mouse["button_cache"][num] or multiple_click:
                    _temp.Mouse["button_cache"][num] = _temp.Mouse["button_cache"][num] if multiple_click else True
                    return True
        if not MouseState(-1)[num]:
            _temp.Mouse["button_cache"][num] = False
    return False

def MouseScroll():
    for event in AllEvents():
        if event.type == sdl2.SDL_MOUSEWHEEL:
            return round(event.wheel.preciseY)
    return 0

def simpleHover(window, location, size):
    location = (location[0] + window.location[0], location[1] + window.location[1]) if window.__WindowId__ < 0 else location
    _position = GetMousePosition()
    return location[0] < _position[0] < location[0] + size[0] and location[1] < _position[1] < location[1] + size[1]

def object2dSimpleHover(window, camera, location, size, scale, pivot):
    _render, cache_location, cache_size = _calculate_object2d(location, size, 0, scale, window, camera, pivot)
    return _render and simpleHover(window, cache_location, cache_size)



import os
from threading import Thread

from ._sdl import sdlmixer
from .Log import LogError

__all__ = ["Sound", "Music"]

class Sound:
    def __init__(self, path: str, preload: bool = True):
        self._path = path
        self.sound = None
        self._channel = 0
        self._volume = 100
        self._load_type = 0
        if preload:
            self.load()

    def _threading_load(self, path: str | None):
        _threading = Thread(target=self._load, args=(path, ))
        _threading.daemon = True
        _threading.start()

    def _load(self, path: str | None):
        self._path = path if path is not None else self._path
        if os.path.exists(self._path):
            self.sound = sdlmixer.Mix_LoadWAV(self._path.encode())
        if not self.sound:
            LogError(f"Unable to load sound '{self._path}'")

    def load(self, path: str | None = None):
        if self._load_type == 0:
            self._threading_load(path)
        elif self._load_type == 1:
            self._load(path)

    def play(self, loops: int = -1):
        sdlmixer.Mix_PlayChannel(self._channel, self.sound, loops)

    def pause(self):
        sdlmixer.Mix_Pause(self._channel)

    def resume(self):
        sdlmixer.Mix_Resume(self._channel)

    @property
    def volume(self):
        return self._volume

    @volume.setter
    def volume(self, volume: int):
        self._volume = volume
        sdlmixer.Mix_VolumeChunk(self.sound, self._volume)

class Music:
    def __init__(self, path: str):
        if os.path.exists(path):
            self.music = sdlmixer.Mix_LoadMUS(path.encode())
        else:
            LogError(f"Unable to load sound '{path}'")
        self._volume = 100

    def play(self, loops: int = -1):
        sdlmixer.Mix_PlayMusic(self.music, loops)

    def pause(self):
        sdlmixer.Mix_PauseMusic()

    def resume(self):
        sdlmixer.Mix_ResumeMusic()

    @property
    def position(self):
        return round(sdlmixer.Mix_GetMusicPosition(self.music))

    @position.setter
    def position(self, position: int):
        sdlmixer.Mix_SetMusicPosition(position)

    @property
    def volume(self):
        return self._volume

    @volume.setter
    def volume(self, volume: int):
        self._volume = volume
        sdlmixer.Mix_VolumeMusic(self._volume)



from time import time, sleep

__all__ = ["Time", "fps_to_time", "get_fps_from_time"]

class Time:
    def __init__(self, delta_time: int | float):
        self._tick_time = time()
        self._delta_time = delta_time
        self._elapsed_time = time() - self._tick_time
        self._flipflop_bool = False

    @property
    def delta_time(self) -> int | float:
        return self._delta_time

    @delta_time.setter
    def delta_time(self, delta_time: int | float):
        self._delta_time = delta_time

    @property
    def flipflop(self):
        return self._flipflop_bool

    @flipflop.setter
    def flipflop(self, b):
        self._flipflop_bool = bool(b)

    def restart(self):
        self._tick_time = time()

    def tick(self, restart=False) -> bool:
        self._elapsed_time = time() - self._tick_time
        if self._elapsed_time >= self._delta_time:
            if restart:
                self.restart()
            return True
        return False

    def tickSleep(self):
        self._elapsed_time = time() - self._tick_time
        if self._elapsed_time < self._delta_time:
            sleep(self._delta_time - self._elapsed_time)
        self.restart()

    def tickMotion(self) -> float:
        self._elapsed_time = time() - self._tick_time
        self.restart()
        return 1 if self._delta_time == 0 else 0.01 * ((self._elapsed_time / self._delta_time) * 100)

    def tickFlipFlop(self) -> bool:
        self._elapsed_time = time() - self._tick_time
        if self._elapsed_time >= self._delta_time:
            self.restart()
            self._flipflop_bool = not self._flipflop_bool
        return self._flipflop_bool

def fps_to_time(target_fps):
    return 0 if target_fps == 0 else 1.0 / target_fps

def get_fps_from_time(object_time: Time) -> int:
    return round(1 / object_time._elapsed_time if not object_time._elapsed_time == 0 else 1 / (object_time._delta_time if object_time._delta_time > 0 else 1))



from ctypes import c_int, c_uint, c_long, c_char_p, c_void_p, byref, POINTER as _P
from ._dll_loader import TTFFunc
from .sdl2 import Uint16, Uint32, SDL_RWops, SDL_version, SDL_Color, SDL_Surface

TTF_STYLE_NORMAL = 0x00
TTF_STYLE_BOLD = 0x01
TTF_STYLE_ITALIC = 0x02
TTF_STYLE_UNDERLINE = 0x04
TTF_STYLE_STRIKETHROUGH = 0x08

TTF_Direction = c_int
TTF_DIRECTION_LTR = 0
TTF_DIRECTION_RTL = 1
TTF_DIRECTION_TTB = 2
TTF_DIRECTION_BTT = 3

hb_direction_t = c_int
HB_DIRECTION_INVALID = 0
HB_DIRECTION_LTR = 4
HB_DIRECTION_RTL = 5
HB_DIRECTION_TTB = 6
HB_DIRECTION_BTT = 7

class TTF_Font(c_void_p):
    pass

def TTF_Linked_Version(): return TTFFunc("TTF_Linked_Version", None, _P(SDL_version))().contents
def TTF_GetFreeTypeVersion(major, minor, patch): return TTFFunc("TTF_GetFreeTypeVersion", [_P(c_int), _P(c_int), _P(c_int)])(major, minor, patch)

def TTF_Init(): return TTFFunc("TTF_Init", None, c_int)()
def TTF_WasInit(): return TTFFunc("TTF_WasInit", None, c_int)()
def TTF_Quit(): return TTFFunc("TTF_Quit")()

def TTF_OpenFont(file, ptsize) -> TTF_Font: return TTFFunc("TTF_OpenFont", [c_char_p, c_int], _P(TTF_Font))(file.encode(), ptsize)
def TTF_OpenFontDPI(file, ptsize, hdpi, vdpi) -> TTF_Font: return TTFFunc("TTF_OpenFontDPI", [c_char_p, c_int, c_uint, c_uint], _P(TTF_Font))(file, ptsize, hdpi, vdpi)
def TTF_OpenFontRW(src, freesrc, ptsize) -> TTF_Font: return TTFFunc("TTF_OpenFontRW", [_P(SDL_RWops), c_int, c_int], _P(TTF_Font))(src, freesrc, ptsize)
def TTF_OpenFontDPIRW(src, freesrc, ptsize, hdpi, vdpi) -> TTF_Font: return TTFFunc("TTF_OpenFontDPIRW", [_P(SDL_RWops), c_int, c_int, c_uint, c_uint], _P(TTF_Font))(src, freesrc, ptsize, hdpi, vdpi)
def TTF_CloseFont(font: TTF_Font): return TTFFunc("TTF_CloseFont", [_P(TTF_Font)])(font)

def TTF_SetFontSize(font: TTF_Font, ptsize): return TTFFunc("TTF_SetFontSize", [_P(TTF_Font), c_int], c_int)(font, ptsize)
def TTF_SetFontSizeDPI(font: TTF_Font, ptsize, hdpi, vdpi): return TTFFunc("TTF_SetFontSizeDPI", [_P(TTF_Font), c_int, c_uint, c_uint], c_int)(font, ptsize, hdpi, vdpi)

def TTF_GetFontStyle(font: TTF_Font): return TTFFunc("TTF_GetFontStyle", [_P(TTF_Font)], c_int)(font)
def TTF_SetFontStyle(font: TTF_Font, style): return TTFFunc("TTF_SetFontStyle", [_P(TTF_Font), c_int], None)(font, style)

def TTF_GetFontOutline(font: TTF_Font): return TTFFunc("TTF_GetFontOutline", [_P(TTF_Font)], c_int)(font)
def TTF_SetFontOutline(font: TTF_Font, outline): return TTFFunc("TTF_SetFontOutline", [_P(TTF_Font), c_int], None)(font, outline)

def TTF_FontHeight(font: TTF_Font): return TTFFunc("TTF_FontHeight", [_P(TTF_Font)], c_int)(font)
def TTF_FontAscent(font: TTF_Font): return TTFFunc("TTF_FontAscent", [_P(TTF_Font)], c_int)(font)
def TTF_FontDescent(font: TTF_Font): return TTFFunc("TTF_FontDescent", [_P(TTF_Font)], c_int)(font)
def TTF_FontLineSkip(font: TTF_Font): return TTFFunc("TTF_FontLineSkip", [_P(TTF_Font)], c_int)(font)

def TTF_GetFontKerning(font: TTF_Font): return TTFFunc("TTF_GetFontKerning", [_P(TTF_Font)], c_int)(font)
def TTF_SetFontKerning(font: TTF_Font, allowed): return TTFFunc("TTF_SetFontKerning", [_P(TTF_Font), c_int])(font, allowed)

def TTF_FontFaces(font: TTF_Font): return TTFFunc("TTF_FontFaces", [_P(TTF_Font)], c_long)(font)
def TTF_FontFaceIsFixedWidth(font: TTF_Font): return TTFFunc("TTF_FontFaceIsFixedWidth", [_P(TTF_Font)], c_int)(font)
def TTF_FontFaceFamilyName(font: TTF_Font): return TTFFunc("TTF_FontFaceFamilyName", [_P(TTF_Font)], c_char_p)(font)
def TTF_FontFaceStyleName(font: TTF_Font): return TTFFunc("TTF_FontFaceStyleName", [_P(TTF_Font)], c_char_p)(font)

def TTF_GlyphIsProvided(font: TTF_Font, ch): return TTFFunc("TTF_GlyphIsProvided", [_P(TTF_Font), Uint16], c_int)(font, ch)
def TTF_GlyphIsProvided32(font: TTF_Font, ch): return TTFFunc("TTF_GlyphIsProvided32", [_P(TTF_Font), Uint32], c_int)(font, ch)

def TTF_SizeText(font: TTF_Font, text) -> tuple:
    w, h = c_long(0), c_long(0)
    TTFFunc("TTF_SizeText", [_P(TTF_Font), c_char_p, _P(c_int), _P(c_int)], c_int)(font, text, byref(w), byref(h))
    return w.value, h.value

def TTF_SizeUTF8(font: TTF_Font, text) -> tuple:
    w, h = c_long(0), c_long(0)
    TTFFunc("TTF_SizeUTF8", [_P(TTF_Font), c_char_p, _P(c_int), _P(c_int)], c_int)(font, text, byref(w), byref(h))
    return w.value, h.value

def TTF_RenderText_Shaded(font: TTF_Font, text, fg, bg) -> SDL_Surface: return TTFFunc("TTF_RenderText_Shaded", [_P(TTF_Font), c_char_p, SDL_Color, SDL_Color], _P(SDL_Surface))(font, text, fg, bg)
def TTF_RenderUTF8_Shaded(font: TTF_Font, text, fg, bg) -> SDL_Surface: return TTFFunc("TTF_RenderUTF8_Shaded", [_P(TTF_Font), c_char_p, SDL_Color, SDL_Color], _P(SDL_Surface))(font, text, fg, bg)

def TTF_RenderText_Shaded_Wrapped(font: TTF_Font, text, fg, bg, wrapLength) -> SDL_Surface: return TTFFunc("TTF_RenderText_Shaded_Wrapped", [_P(TTF_Font), c_char_p, SDL_Color, SDL_Color, Uint32], _P(SDL_Surface))(font, text, fg, bg, wrapLength)
def TTF_RenderUTF8_Shaded_Wrapped(font: TTF_Font, text, fg, bg, wrapLength) -> SDL_Surface: return TTFFunc("TTF_RenderUTF8_Shaded_Wrapped", [_P(TTF_Font), c_char_p, SDL_Color, SDL_Color, Uint32], _P(SDL_Surface))(font, text, fg, bg, wrapLength)

def TTF_RenderText_Blended(font: TTF_Font, text, fg) -> SDL_Surface: return TTFFunc("TTF_RenderText_Blended", [_P(TTF_Font), c_char_p, SDL_Color], _P(SDL_Surface))(font, text.encode(), fg).contents
def TTF_RenderUTF8_Blended(font: TTF_Font, text, fg) -> SDL_Surface: return TTFFunc("TTF_RenderUTF8_Blended", [_P(TTF_Font), c_char_p, SDL_Color], _P(SDL_Surface))(font, text, fg)

def TTF_RenderText_Blended_Wrapped(font: TTF_Font, text, fg, wrapLength) -> SDL_Surface: return TTFFunc("TTF_RenderText_Blended_Wrapped", [_P(TTF_Font), c_char_p, SDL_Color, Uint32], _P(SDL_Surface))(font, text, fg, wrapLength)
def TTF_RenderUTF8_Blended_Wrapped(font: TTF_Font, text, fg, wrapLength) -> SDL_Surface: return TTFFunc("TTF_RenderUTF8_Blended_Wrapped", [_P(TTF_Font), c_char_p, SDL_Color, Uint32], _P(SDL_Surface))(font, text, fg, wrapLength)

def TTF_SetScript(script): return TTFFunc("TTF_SetScript", [c_int], c_int)(script)
def TTF_SetDirection(direction): return TTFFunc("TTF_SetDirection", [c_int], c_int)(direction)
def TTF_SetFontDirection(font: TTF_Font, direction): return TTFFunc("TTF_SetFontDirection", [_P(TTF_Font), TTF_Direction], c_int)(font, direction)

def TTF_SetFontScriptName(font: TTF_Font, script): return TTFFunc("TTF_SetFontScriptName", [_P(TTF_Font), c_char_p], c_int)(font, script)
def TTF_GetFontKerningSize(font: TTF_Font, prev_index, index): return TTFFunc("TTF_GetFontKerningSize", [_P(TTF_Font), c_int, c_int], c_int)(font, prev_index, index)



from .Mesh import Mesh2D
from .Color import Color
from ._sdl.sdl2 import SDL_PIXELFORMAT_RGB24, SDL_PIXELFORMAT_ARGB4444, SDL_PIXELFORMAT_RGBA8888, SDL_PIXELFORMAT_ARGB8888

__all__ = ["WindowFlag", "WindowEvent", "RenderDriver", "ImageFormat",
           "ControllerType", "ControllerButton",
           "ConstMouseButton", "KeyboardButton",
           "Colors",
           "Vector", "Pivot2D", "Meshes2D", "All"]

All = -1

class WindowFlag:
    Fullscreen = 0X00000001
    Opengl = 0X00000002
    Shown = 0X00000004
    Hidden = 0X00000008
    Borderless = 0X00000010
    Resizable = 0X00000020
    Minimized = 0X00000040
    Maximized = 0X00000080
    MouseGrabbed = 0X00000100
    InputFocus = 0X00000200
    MouseFocus = 0X00000400
    FullscreenDesktop = (Fullscreen | 0X00001000)
    Foreign = 0X00000800
    AllowHighdpi = 0X00002000
    MouseCapture = 0X00004000
    AlwaysOnTop = 0X00008000
    SkipTaskbar = 0X00010000
    Utility = 0X00020000
    Tooltip = 0X00040000
    PopupMenu = 0X00080000
    KeyboardGrabbed = 0X00100000
    #Vulkan = 0X10000000
    #Metal = 0X20000000

class WindowEvent:
    NONE = 0
    SHOWN = 1
    HIDDEN = 2
    EXPOSED = 3
    MOVED = 4
    RESIZED = 5
    SIZE_CHANGED = 6
    MINIMIZED = 7
    MAXIMIZED = 8
    RESTORED = 9
    ENTER = 10
    LEAVE = 11
    FOCUS_GAINED = 12
    FOCUS_LOST = 13
    CLOSE = 14
    TAKE_FOCUS = 15
    HIT_TEST = 16
    ICCPROF_CHANGED = 17
    DISPLAY_CHANGED = 18

class RenderDriver:
    Direct3d = All
    Direct3d11 = All
    Direct3d12 = All
    OpenGL = All
    Software = All

class ImageFormat:
    RGB = (24, SDL_PIXELFORMAT_RGB24)
    ARGB = (32, SDL_PIXELFORMAT_ARGB8888)
    ARGB16 = (16, SDL_PIXELFORMAT_ARGB4444)
    RGBA = (32, SDL_PIXELFORMAT_RGBA8888)

class ControllerType:
    Unknown = 0
    Xbox360 = 1
    XboxONE = 2
    Ps3 = 3
    Ps4 = 4
    NintendoSwitchPRO = 5
    Virtual = 6
    Ps5 = 7
    AmazonLuna = 8
    GoogleStadia = 9
    NvidiaShield = 10
    NintendoSwitchJoyconLeft = 11
    NintendoSwitchJoyconRight = 12
    NintendoSwitchJoyconPair = 13

class ControllerButton:
    A = 0
    B = 1
    X = 2
    Y = 3
    Back = 4
    Guide = 5
    Start = 6
    LeftStick = 7
    RightStick = 8
    LeftShoulder = 9
    RightShoulder = 10
    Up = 11
    Down = 12
    Left = 13
    Right = 14
    Misc1 = 15
    Paddle1 = 16
    Paddle2 = 17
    Paddle3 = 18
    Paddle4 = 19
    Touchpad = 20

class ConstMouseButton:
    Left = 1
    Middle = 2
    Right = 3
    X1 = 4
    X2 = 5

class KeyboardButton:
    KeyA = 4
    KeyB = 5
    KeyC = 6
    KeyD = 7
    KeyE = 8
    KeyF = 9
    KeyG = 10
    KeyH = 11
    KeyI = 12
    KeyJ = 13
    KeyK = 14
    KeyL = 15
    KeyM = 16
    KeyN = 17
    KeyO = 18
    KeyP = 19
    KeyQ = 20
    KeyR = 21
    KeyS = 22
    KeyT = 23
    KeyU = 24
    KeyV = 25
    KeyW = 26
    KeyX = 27
    KeyY = 28
    KeyZ = 29

    #KeyCedilla = 51

    Key1 = 30
    Key2 = 31
    Key3 = 32
    Key4 = 33
    Key5 = 34
    Key6 = 35
    Key7 = 36
    Key8 = 37
    Key9 = 38
    Key0 = 39

    Return = 40
    Esc = 41
    Back = 42
    Tab = 43
    Space = 44

    Minus = 45
    Equals = 46

    F1 = 58
    F2 = 59
    F3 = 60
    F4 = 61
    F5 = 62
    F6 = 63
    F7 = 64
    F8 = 65
    F9 = 66
    F10 = 67
    F11 = 68
    F12 = 69

    Right = 79
    Left = 80
    Down = 81
    Up = 82

    LeftShift = 225
    RightShift = 229
    Shift = (LeftShift, RightShift)
    LeftCtrl = 224
    RightCtrl = 228
    Ctrl = (LeftCtrl, RightCtrl)
    LeftAlt = 226
    RightAlt = 230
    Alt = (LeftAlt, RightAlt)

class Colors:
    StandardColor = Color(120, 120, 255)

    # Primary colors
    Red = Color(255, 0, 0)
    Green = Color(0, 255, 0)
    Blue = Color(0, 0, 255)

    # Secondary colors
    Yellow = Color(255, 255, 0)
    Cyan = Color(0, 255, 255)
    Magenta = Color(255, 0, 255)

    # Shades of black and white
    Black = Color(0, 0, 0)
    White = Color(255, 255, 255)
    Gray = Color(128, 128, 128)
    DarkGray = Color(64, 64, 64)
    LightGray = Color(192, 192, 192)

    # Pastel colors
    PastelRed = Color(255, 182, 193)
    PastelGreen = Color(152, 251, 152)
    PastelBlue = Color(173, 216, 230)
    PastelYellow = Color(255, 255, 204)
    PastelPurple = Color(216, 191, 216)
    PastelPink = Color(255, 182, 193)

    # Bright colors
    NeonGreen = Color(57, 255, 20)
    NeonBlue = Color(77, 77, 255)
    NeonPink = Color(255, 20, 147)
    NeonYellow = Color(255, 255, 0)
    NeonOrange = Color(255, 165, 0)

    # Neutral colors
    Beige = Color(245, 245, 220)
    Ivory = Color(255, 255, 240)
    Tan = Color(210, 180, 140)
    Khaki = Color(240, 230, 140)
    Lavender = Color(230, 230, 250)
    Mint = Color(189, 252, 201)
    SlateGray = Color(112, 128, 144)
    LightSlateGray = Color(119, 136, 153)

    # Warm colors
    Orange = Color(255, 165, 0)
    DarkOrange = Color(255, 140, 0)
    Coral = Color(255, 127, 80)
    Tomato = Color(255, 99, 71)
    Salmon = Color(250, 128, 114)
    Firebrick = Color(178, 34, 34)
    Brown = Color(165, 42, 42)
    Maroon = Color(128, 0, 0)

    # Cold colors
    Navy = Color(0, 0, 128)
    Teal = Color(0, 128, 128)
    Turquoise = Color(64, 224, 208)
    Aqua = Color(0, 255, 255)
    SkyBlue = Color(135, 206, 235)
    RoyalBlue = Color(65, 105, 225)
    SteelBlue = Color(70, 130, 180)
    MidnightBlue = Color(25, 25, 112)

    # Metallic colors
    Gold = Color(255, 215, 0)
    Silver = Color(192, 192, 192)
    Bronze = Color(205, 127, 50)

    # Natural colors
    ForestGreen = Color(34, 139, 34)
    LimeGreen = Color(50, 205, 50)
    Olive = Color(128, 128, 0)
    DarkOliveGreen = Color(85, 107, 47)
    SeaGreen = Color(46, 139, 87)

    # Skin colors
    Peach = Color(255, 218, 185)
    Bisque = Color(255, 228, 196)
    Wheat = Color(245, 222, 179)
    RosyBrown = Color(188, 143, 143)

    # other colors
    Indigo = Color(75, 0, 130)
    Violet = Color(238, 130, 238)
    Plum = Color(221, 160, 221)
    Crimson = Color(220, 20, 60)
    Fuchsia = Color(255, 0, 255)
    Chartreuse = Color(127, 255, 0)
    Periwinkle = Color(204, 204, 255)

#class Themes:
#    Standard = {}

class Vector:
    X = 1
    Y = 2
    GLOBAL = 10
    LOCAL = 30

class Pivot2D:
    Center = 800
    TopLeftSide = 700
    TopRightSide = 750
    LowerLeftSide = 600
    LowerRightSide = 650

class Meshes2D:
    Plane = Mesh2D([])
    Star = Mesh2D([(150, 0), (186, 116), (300, 116), (212, 192), (250, 300), (150, 230), (50, 300), (88, 192), (0, 116), (114, 116)])
    Arrow = Mesh2D([(0, 100), (30, 100), (30, 300), (38, 300), (38, 100), (68, 100), (34, 0)])



from setuptools import setup

setup(
    name="MGE",
    version="1.0.0.8-beta",
    license='zlib',
    description="LibMGE is a graphical user interface library for developing 2D programs and games.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Lucas Guimarães",
    author_email="commercial@lucasguimaraes.pro",
    url="https://libmge.org/",
    project_urls={
        "Source": "https://github.com/MonumentalGames/LibMGE",
        "Documentation": "https://docs.libmge.org/",
        "Author Website": "https://lucasguimaraes.pro/"
    },
    python_requires=">=3.5",
    packages=[
          "MGE",
          "MGE/_sdl",
          "MGE/_InputsEmulator",
          "MGE/_ConsoleScripts"
    ],
    install_requires=[
        "numpy"
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        'License :: OSI Approved :: zlib/libpng License',
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Operating System :: Microsoft :: Windows",
    ],
    entry_points={
        'console_scripts': [
            'MGE=MGE._ConsoleScripts:main',
            'mge=MGE._ConsoleScripts:main',
        ],
    },
    keywords="2D development, graphical interface, games",
)



import os
from ctypes import CDLL

from ..Log import LogError, LogCritical

def NullFunction(*args):
    return

def find_path(file_name: str | list):
    file_name = file_name if type(file_name) == str else file_name[0]
    if os.path.exists(f"{os.path.abspath('./')}/{file_name}"):
        dll_path = os.path.abspath(f'./{file_name}')
    elif os.path.exists(f"{os.path.abspath(os.path.dirname(__file__))}/../{file_name}"):
        dll_path = f"{os.path.abspath(os.path.dirname(__file__))}/../{file_name}"
    else:
        LogCritical(f"{file_name} not found", "", 1)
    return dll_path

class DLL(object):
    def __init__(self, path=None):
        self._dll = None if path is None else CDLL(path)

    def bind_function(self, func_name, args=None, returns=None):
        if self._dll is None:
            return NullFunction
        else:
            func = getattr(self._dll, func_name, None)
            if not func:
                LogError(f"{func_name}")
                return NullFunction
            func.argtypes, func.restype = args, returns
            return func



import ctypes

__all__ = ["MousePress", "MouseRelease", "MouseClick", "MouseScroll",
           "KeyboardPress", "KeyboardRelease", "KeyboardClick"]

_MouseEvent = {1: [0x0002, 0x0004], 2: [0x0020, 0x0040], 3: [0x0008, 0x0010], 4: [0x0002, 0x0004], 5: [0x0002, 0x0004]}
_KeyboardEvent = {}

def MousePress(button):
    pass

def MouseRelease(button):
    pass

def MouseClick(button):
    pass

def MouseScroll(scroll):
    pass

def KeyboardPress(key):
    pass

def KeyboardRelease(key):
    pass

def KeyboardClick(key):
    pass



from .Time import Time, fps_to_time

__all__ = ["Camera"]

class Camera:
    def __init__(self, location=(0, 0)):
        self._location = list(location)
        self.motion_tick_time = {"x": Time(fps_to_time(60)), "y": Time(fps_to_time(60))}

    def motionTimeStart(self, axis=-1):
        if axis == 1 or axis == -1:  # x
            self.motion_tick_time["x"].restart()
        if axis == 2 or axis == -1:  # y
            self.motion_tick_time["y"].restart()

    def motion(self, axis, speed):
        if axis == 1 or axis == -1:  # x
            self._location[0] += (speed if type(speed) == int or type(speed) == float else speed[0]) * self.motion_tick_time["x"].tickMotion()
        if axis == 2 or axis == -1:  # y
            self._location[1] += (speed if type(speed) == int or type(speed) == float else speed[1]) * self.motion_tick_time["y"].tickMotion()

    @property
    def location(self):
        return [int(self._location[0]), int(self._location[1])]

    @location.setter
    def location(self, location):
        self._location = location



"""
LibMGE is an open-source library created for the development of 2D programs and games in Python.
Our goal is to provide a simple and accessible solution, allowing developers,
from beginners to professionals, to rely on an efficient tool for creating
games and graphical interfaces.

Whether for prototyping or full game development, LibMGE was designed to make the process more
intuitive and faster without compromising quality.

zlib License

Copyright (c) 2024 Lucas Guimarães

This software is provided 'as-is', without any express or implied
warranty.  In no event will the authors be held liable for any damages
arising from the use of this software.

Permission is granted to anyone to use this software for any purpose,
including commercial applications, and to alter it and redistribute it
freely, subject to the following restrictions:

1. The origin of this software must not be misrepresented; you must not
   claim that you wrote the original software. If you use this software
   in a product, an acknowledgment in the product documentation would be
   appreciated but is not required.
2. Altered source versions must be plainly marked as such, and must not be
   misrepresented as being the original software.
3. This notice may not be removed or altered from any source distribution.
"""
import sys

if not sys.argv or sys.argv[0] != f"{sys.prefix}\\Scripts\\mge.exe":
    from .Common import update, init, SetLogicClock, GetLogicClock, AllEvents, QuitEvent, WindowEvents, OpenUrl, AutoCalcs2D
    from .Log import *
    from .Audio import Music, Sound
    from .Platform import Platform
    from .Constants import *
    from .Camera import Camera
    from .Window import Window, CreateGlWindow, InternalWindow
    from .Monitors import Monitors, Monitor
    from .Text import ObjectText, ObjectInputTextLine, ObjectInputPassLine, ObjectInputTextBox
    from .Object2D import Object2D
    from .Line import Line
    from .Material import Material
    from .Texture import Texture
    from .Image import *
    from .Color import Color
    from .Mesh import Mesh2D, CreateMeshPlane
    from .Mouse import *
    from .Keyboard import keyboard
    from .GameController import *
    from .Button import *
    from .Time import *
    from .Version import __version__, __versionData__, __versionList__

    print(f"{'MGE' if __versionData__['phase'] == 'Stable' else __versionData__['phase'] + '-MGE'} {__version__} (SDL {Platform.SDL.SDL_version} | Python {Platform.python_version})")



from ctypes import Structure, CFUNCTYPE, c_int, c_char_p, c_void_p, c_double, POINTER as _P
from ._dll_loader import MIXERFunc
from .sdl2 import Uint8, Uint16, Uint32, Sint16, SDL_bool, SDL_RWops, SDL_RWFromFile, SDL_version

Mix_Fading = c_int
MIX_NO_FADING = 0
MIX_FADING_OUT = 1
MIX_FADING_IN = 2

class Mix_Chunk(Structure):
    _fields_ = [("allocated", c_int), ("abuf", _P(Uint8)), ("alen", Uint32), ("volume", Uint8)]

class Mix_Music(c_void_p):
    pass

Mix_EffectFunc_t = CFUNCTYPE(None, c_int, c_void_p, c_int, c_void_p)
Mix_EffectDone_t = CFUNCTYPE(None, c_int, c_void_p)
soundfont_function = CFUNCTYPE(c_int, c_char_p, c_void_p)

def Mix_Linked_Version(): return MIXERFunc("Mix_Linked_Version", None, _P(SDL_version))().contents

def Mix_Init(flags=0): return MIXERFunc("Mix_Init", [c_int], c_int)(flags)
def Mix_Quit(): return MIXERFunc("Mix_Quit")()

def Mix_OpenAudio(frequency, format, channels, chunksize): return MIXERFunc("Mix_OpenAudio", [c_int, Uint16, c_int, c_int], c_int)(frequency, format, channels, chunksize)
def Mix_OpenAudioDevice(frequency, format, channels, chunksize, device, allowed_changes): return MIXERFunc("Mix_OpenAudioDevice", [c_int, Uint16, c_int, c_int, c_char_p, c_int], c_int)(frequency, format, channels, chunksize, device, allowed_changes)

def Mix_AllocateChannels(numchans): return MIXERFunc("Mix_AllocateChannels", [c_int], c_int)(numchans)

def Mix_QuerySpec(frequency, format, channels): return MIXERFunc("Mix_QuerySpec", [_P(c_int), _P(Uint16), _P(c_int)], c_int)(frequency, format, channels)

def Mix_LoadWAV_RW(src, freesrc): return MIXERFunc("Mix_LoadWAV_RW", [_P(SDL_RWops), c_int], _P(Mix_Chunk))(src, freesrc)
def Mix_LoadMUS_RW(src, freesrc): return MIXERFunc("Mix_LoadMUS_RW", [_P(SDL_RWops)], _P(Mix_Music))(src, freesrc)

def Mix_LoadWAV(file): return Mix_LoadWAV_RW(SDL_RWFromFile(file, b"rb"), 1)
def Mix_LoadMUS(file): return MIXERFunc("Mix_LoadMUS", [c_char_p], _P(Mix_Music))(file)

def Mix_QuickLoad_WAV(mem): return MIXERFunc("Mix_QuickLoad_WAV", [_P(Uint8)], _P(Mix_Chunk))(mem)
def Mix_QuickLoad_RAW(mem, len): return MIXERFunc("Mix_QuickLoad_RAW", [_P(Uint8), Uint32], _P(Mix_Chunk))(mem, len)

def Mix_FreeChunk(chunk): return MIXERFunc("Mix_FreeChunk", [_P(Mix_Chunk)])(chunk)
def Mix_FreeMusic(music): return MIXERFunc("Mix_FreeMusic", [_P(Mix_Music)])(music)

def Mix_GetNumChunkDecoders(): return MIXERFunc("Mix_GetNumChunkDecoders", None, c_int)()
def Mix_GetChunkDecoder(index): return MIXERFunc("Mix_GetChunkDecoder", [c_int], c_char_p)(index)
def Mix_HasChunkDecoder(name): return MIXERFunc("Mix_HasChunkDecoder", [c_char_p], SDL_bool)(name)
def Mix_GetNumMusicDecoders(): return MIXERFunc("Mix_GetNumMusicDecoders", None, c_int)()
def Mix_GetMusicDecoder(index): return MIXERFunc("Mix_GetMusicDecoder", [c_int], c_char_p)(index)
def Mix_HasMusicDecoder(name): return MIXERFunc("Mix_HasMusicDecoder", [c_char_p], SDL_bool)(name)
def Mix_GetMusicTitle(music): return MIXERFunc("Mix_GetMusicTitle", [_P(Mix_Music)], c_char_p)(music)
def Mix_GetMusicTitleTag(music): return MIXERFunc("Mix_GetMusicTitleTag", [_P(Mix_Music)], c_char_p)(music)
def Mix_GetMusicArtistTag(music): return MIXERFunc("Mix_GetMusicArtistTag", [_P(Mix_Music)], c_char_p)(music)
def Mix_GetMusicAlbumTag(music): return MIXERFunc("Mix_GetMusicAlbumTag", [_P(Mix_Music)], c_char_p)(music)
def Mix_GetMusicCopyrightTag(music): return MIXERFunc("Mix_GetMusicCopyrightTag", [_P(Mix_Music)], c_char_p)(music)
def Mix_SetPostMix(mix_func, arg): return MIXERFunc("Mix_SetPostMix", [mix_func, c_void_p])(mix_func, arg)
def Mix_HookMusic(mix_func, arg): return MIXERFunc("Mix_HookMusic", [mix_func, c_void_p])(mix_func, arg)
def Mix_HookMusicFinished(music_finished): return MIXERFunc("Mix_HookMusicFinished", [music_finished])(music_finished)
def Mix_GetMusicHookData(): return MIXERFunc("Mix_GetMusicHookData", None, c_void_p)()
def Mix_ChannelFinished(channel_finished): return MIXERFunc("Mix_ChannelFinished", [channel_finished])(channel_finished)
def Mix_RegisterEffect(chan, f, d, arg): return MIXERFunc("Mix_RegisterEffect", [c_int, Mix_EffectFunc_t, Mix_EffectDone_t, c_void_p], c_int)(chan, f, d, arg)
def Mix_UnregisterEffect(channel, f): return MIXERFunc("Mix_UnregisterEffect", [c_int, Mix_EffectFunc_t], c_int)(channel, f)
def Mix_UnregisterAllEffects(channel): return MIXERFunc("Mix_UnregisterAllEffects", [c_int])(channel)
def Mix_SetPanning(channel, left, right): return MIXERFunc("Mix_SetPanning", [c_int, Uint8, Uint8], c_int)(channel, left, right)
def Mix_SetPosition(channel, angle, distance): return MIXERFunc("Mix_SetPosition", [c_int, Sint16, Uint8], c_int)(channel, angle, distance)
def Mix_SetDistance(channel, distance): return MIXERFunc("Mix_SetDistance", [c_int, Uint8])(channel, distance)
def Mix_SetReverseStereo(channel, flip): return MIXERFunc("Mix_SetReverseStereo", [c_int, c_int], c_int)(channel, flip)
def Mix_ReserveChannels(num): return MIXERFunc("Mix_ReserveChannels", [c_int], c_int)(num)

def Mix_GroupChannel(which, tag): return MIXERFunc("Mix_GroupChannel", [c_int, c_int], c_int)(which, tag)
def Mix_GroupChannels(from_, to, tag): return MIXERFunc("Mix_GroupChannels", [c_int, c_int, c_int], c_int)(from_, to, tag)
def Mix_GroupAvailable(tag): return MIXERFunc("Mix_GroupAvailable", [c_int], c_int)(tag)
def Mix_GroupCount(tag): return MIXERFunc("Mix_GroupCount", [c_int], c_int)(tag)
def Mix_GroupOldest(tag): return MIXERFunc("Mix_GroupOldest", [c_int], c_int)(tag)
def Mix_GroupNewer(tag): return MIXERFunc("Mix_GroupNewer", [c_int], c_int)(tag)

def Mix_PlayChannelTimed(channel, chunk, loops, ticks): return MIXERFunc("Mix_PlayChannelTimed", [c_int, _P(Mix_Chunk), c_int, c_int], c_int)(channel, chunk, loops, ticks)
def Mix_PlayChannel(channel, chunk, loops): return MIXERFunc("Mix_PlayChannel", [c_int, _P(Mix_Chunk), c_int], c_int)(channel, chunk, loops)
def Mix_PlayMusic(music, loops): return MIXERFunc("Mix_PlayMusic", [_P(Mix_Music), c_int], c_int)(music, loops)

def Mix_FadeInMusic(music, loops, ms): return MIXERFunc("Mix_FadeInMusic", [_P(Mix_Music), c_int, c_int], c_int)(music, loops, ms)
def Mix_FadeInMusicPos(music, loops, ms, position): return MIXERFunc("Mix_FadeInMusicPos", [_P(Mix_Music), c_int, c_int, c_double], c_int)(music, loops, ms, position)
def Mix_FadeInChannelTimed(channel, chunk, loops, ms, ticks): return MIXERFunc("Mix_FadeInChannelTimed", [c_int, _P(Mix_Chunk), c_int, c_int, c_int], c_int)(channel, chunk, loops, ms, ticks)
def Mix_FadeInChannel(channel, chunk, loops, ms): return MIXERFunc("Mix_FadeInChannel", [c_int, _P(Mix_Chunk), c_int, c_int], c_int)(channel, chunk, loops, ms)

def Mix_Volume(channel, volume): return MIXERFunc("Mix_Volume", [c_int, c_int], c_int)(channel, volume)
def Mix_VolumeChunk(chunk, volume): return MIXERFunc("Mix_VolumeChunk", [_P(Mix_Chunk), c_int], c_int)(chunk, volume)
def Mix_VolumeMusic(volume): return MIXERFunc("Mix_VolumeMusic", [c_int], c_int)(volume)
def Mix_GetMusicVolume(music): return MIXERFunc("Mix_GetMusicVolume", [_P(Mix_Music)], c_int)(music)
def Mix_MasterVolume(volume): return MIXERFunc("Mix_MasterVolume", [c_int], c_int)(volume)

def Mix_HaltChannel(channel): return MIXERFunc("Mix_HaltChannel", [c_int], c_int)(channel)
def Mix_HaltGroup(tag): return MIXERFunc("Mix_HaltGroup", [c_int], c_int)(tag)
def Mix_HaltMusic(): return MIXERFunc("Mix_HaltMusic", None, c_int)()

def Mix_ExpireChannel(channel, ticks): return MIXERFunc("Mix_ExpireChannel", [c_int, c_int], c_int)(channel, ticks)

def Mix_FadeOutChannel(which, ms): return MIXERFunc("Mix_FadeOutChannel", [c_int, c_int], c_int)(which, ms)
def Mix_FadeOutGroup(tag, ms): return MIXERFunc("Mix_FadeOutGroup", [c_int, c_int], c_int)(tag, ms)
def Mix_FadeOutMusic(ms): return MIXERFunc("Mix_FadeOutMusic", [c_int], c_int)(ms)

def Mix_FadingMusic(): return MIXERFunc("Mix_FadingMusic", None, Mix_Fading)()
def Mix_FadingChannel(which): return MIXERFunc("Mix_FadingChannel", [c_int], Mix_Fading)(which)

def Mix_Pause(channel: int): return MIXERFunc("Mix_Pause", [c_int])(channel)
def Mix_Resume(channel: int): return MIXERFunc("Mix_Resume", [c_int])(channel)
def Mix_Paused(channel: int): return MIXERFunc("Mix_Paused", [c_int], c_int)(channel)
def Mix_PauseMusic(): return MIXERFunc("Mix_PauseMusic")()
def Mix_ResumeMusic(): return MIXERFunc("Mix_ResumeMusic")()
def Mix_RewindMusic(): return MIXERFunc("Mix_RewindMusic")()
def Mix_PausedMusic(): return MIXERFunc("Mix_PausedMusic", None, c_int)()

def Mix_ModMusicJumpToOrder(order): return MIXERFunc("Mix_ModMusicJumpToOrder", [c_int], c_int)(order)

def Mix_SetMusicPosition(position): return MIXERFunc("Mix_SetMusicPosition", [c_double], c_int)(position)
def Mix_GetMusicPosition(music): return MIXERFunc("Mix_GetMusicPosition", [_P(Mix_Music)], c_double)(music)

def Mix_MusicDuration(music): return MIXERFunc("Mix_MusicDuration", [_P(Mix_Music)], c_double)(music)

def Mix_GetMusicLoopStartTime(music): return MIXERFunc("Mix_GetMusicLoopStartTime", [_P(Mix_Music)], c_double)(music)
def Mix_GetMusicLoopEndTime(music): return MIXERFunc("Mix_GetMusicLoopEndTime", [_P(Mix_Music)], c_double)(music)
def Mix_GetMusicLoopLengthTime(music): return MIXERFunc("Mix_GetMusicLoopLengthTime", [_P(Mix_Music)], c_double)(music)

def Mix_Playing(channel): return MIXERFunc("Mix_Playing", [c_int], c_int)(channel)
def Mix_PlayingMusic(): return MIXERFunc("Mix_PlayingMusic", None, c_int)()

def Mix_SetMusicCMD(command): return MIXERFunc("Mix_SetMusicCMD", [c_char_p], c_int)(command)
def Mix_SetSynchroValue(value): return MIXERFunc("Mix_SetSynchroValue", [c_int], c_int)(value)
def Mix_GetSynchroValue(): return MIXERFunc("Mix_GetSynchroValue", None, c_int)()
def Mix_SetSoundFonts(paths): return MIXERFunc("Mix_SetSoundFonts", [c_char_p], c_int)(paths)
def Mix_GetSoundFonts(): return MIXERFunc("Mix_GetSoundFonts", None, c_char_p)()
def Mix_EachSoundFont(function, data): return MIXERFunc("Mix_EachSoundFont", [soundfont_function, c_void_p], c_int)(function, data)
def Mix_SetTimidityCfg(path): return MIXERFunc("Mix_SetTimidityCfg", [c_char_p], c_int)(path)
def Mix_GetTimidityCfg(): return MIXERFunc("Mix_GetTimidityCfg", None, c_char_p)()
def Mix_GetChunk(channel): return MIXERFunc("Mix_GetChunk", [c_int], _P(Mix_Chunk))(channel)
def Mix_CloseAudio(): return MIXERFunc("Mix_CloseAudio")()
